//#define CDI_RS232		// allocate memory

//#include "cdi.def"
//#include "prolog.h"
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#include <process.h>

#include "tinecdi.h"
//#include "dbtools.h"
//#include "cdi.h"	// interface to user program
#include "cdiWinRS232Lib.h"

int busDebugLevel = 0;
int tmoutInterval = 10000;	//ms
int resetInterval = 20000;	//ms

char *ad6150Name = "6151AD";
STRUCT_6150AD enum6150[N_6150_DATA] = {
	{dosisLeistung, cdiDouble, 8},
	{messSize, cdiChar, 1},
	{typInnenRohr, cdiChar, 1},
	{aktiveDetector, cdiChar, 1}
};
STRUCT_DETECTOR_6150AD enumAktiveDetector[] = {
	{0, "Sonde 6150AD-0"},
	{7, "Sonde 6150AD-b"},
	{15, "Sonde 6150AD-15"},
	{17, "Sonde 6150AD-17"},
	{18, "Sonde 6150AD-18"},
	{19, "Sonde 6150AD-19"},
	{20, "Innenzaehlerrohr des 6150AD"},
	{21, "Sonde 6150AD-t Zaehlrohr niedriger Bereich"},
	{22, "Sonde 6150AD-t Zaehlrohr hoher Bereich"},
};


#ifndef CDI_COMM_LINUX
SER_STRUCT serInfo[N_SERIAL_LINES] = {
	{NULL,1}, {NULL,2}, {NULL,3}, {NULL,4}, {NULL,5}, {NULL,6}, {NULL,7}, {NULL,8} 
};

BAUD_STRUCT baudRate[] = {
	{110, CBR_110}, {300, CBR_300}, {600, CBR_600}, {1200, CBR_1200}, {2400, CBR_2400}, {4800, CBR_4800}, {9600, CBR_9600},
  {14400, CBR_14400}, {19200, CBR_19200}, {38400, CBR_38400}, {56000, CBR_56000}, {57600, CBR_57600}, 
  {115200, CBR_115200}, {128000, CBR_128000}, {256000, CBR_256000}
};
#else
SER_STRUCT serInfo[N_SERIAL_LINES] = {
	{-1,1}, {-1,2}, {-1,3}, {-1,4}, {-1,5}, {-1,6}, {-1,7}, {-1,8} 
};

BAUD_STRUCT baudRate[] = {
//	{0, B0}, 
	{50, B50}, {75, B75}, {110, B110}, {134, B134}, {150, B150},{200, B200},{300, B300}, {600, B600},
	{1200, B1200}, {1800, B1800}, {2400, B2400}, {4800, B4800}, {9600, B9600},{19200, B19200},{38400, B38400},
	{57600, B57600},{115200, B115200}
};
#endif
int numBaudRate = sizeof(baudRate) / sizeof(BAUD_STRUCT);

#ifdef CDI_COMM_LINUX
__attribute__((constructor)) void cdiCommMain(void)
{
  int cc = 0;
  if ((cc=cdiRegisterBusFilter(COMM_LIB_NAME,filterRS232Bus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusInitialization(COMM_LIB_NAME,initRS232ForCDI)) != 0 ) goto err;
  if ((cc=cdiRegisterBusHandler(COMM_LIB_NAME,execRS232)) != 0 ) goto err;
  if ((cc=cdiRegisterBusCleanup(COMM_LIB_NAME,closeSerialPort)) != 0 ) goto err;
  if ((cc=cdiRegisterBusReqTimeout(COMM_LIB_NAME,commTimeoutFcn)) != 0 ) goto err;
  RegisterUserCommand("rs232debug",NULL,(int *)&busDebugLevel,NULL,CA_READ|CA_WRITE);

err:
  if (cc)
  {
    cdilog("%s : failure in bus registration %s",erlst[cc],COMM_LIB_NAME);
  }
  return;
}
#endif

CDI_COMM_EXPORT int filterRS232Bus(int *parameter, int accessFlag)
{
  return 0;
}

int get6150ADdatachar(CdiRequestInfoBlk *pReq, char *str, unsigned int leng)
{
  short i, exp, mhi, mlo, mant;
  double dosLeistung;
  char *pText;

  //printf("name1: %s; name2: %s\n",pReq->pDev->lngName, ad6150Name);
  if ( leng < 6 || str[0] != 2 || strcmp(pReq->pDev->lngName, ad6150Name) != 0 ) return 0;
  for(i=0; i<N_6150_DATA; i++)
	if ( pReq->devParameters[index_1] == enum6150[i].index && pReq->pDev->dataFormat == enum6150[i].format && pReq->pDev->dataOuputSize >= enum6150[i].size ) break;
  if ( i == N_6150_DATA )
  {
	pReq->errorCode = non_existent_function;
	if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
	pReq->pDev->numberDeviceError++;
	return 1;
  }
  // Block check 
  for ( exp=str[1], i=2; i<5; i++) exp = exp ^ str[i];
  if ( (char)exp != str[5] )
  {
	//printf("checksum error: 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, check: 0x%x\n",str[1], str[2],str[3],str[4],str[5],exp);
	pReq->errorCode = out_of_sequence;
	if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
	pReq->pDev->numberDeviceError++;
	return 1;
  }
  if ( pReq->devParameters[index_1] == dosisLeistung )
  {
	exp = (int)str[4];
	if ( exp > 127 ) exp -= 256;
	mhi = str[3];
	if ( mhi < 0 ) mhi = 256 + mhi;
	mlo = str[2];
	if ( mlo < 0 ) mlo = 256 + mlo;
	mant = mhi * 256 + mlo;
	dosLeistung = mant * pow(2, exp-15);
	*(double *)pReq->pInData = dosLeistung;
//	printf("data: exp: %d, mhi: %d, mlo: %d, exp: %d, mant: %d, dosis: %f\n",str[4],str[3],str[2],exp,mant,dosLeistung);
	return 1;
  }
  if ( pReq->devParameters[index_1] == messSize )
  {
	pText = (str[1] & 0x80) ? "6150AD in >>/Ex<<-Ausfuehrung(6150ADx/E)" : "sonst(6150ADx oder 6150ADx/H)";
//	  *(short *)pReq->pInData = (short)((str[1] & 0x80) ? 1 : 0);
  }
  else if ( pReq->devParameters[index_1] == typInnenRohr )
  {
	pText = (str[1] & 0x40) ? "ZP1310(6150AD1/3/5)" : "ZP1200(6150AD2/4/6)";
//	  *(short *)pReq->pInData = (short)((str[1] & 0x40) ? 1 : 0);
  }
  else
  {
	exp = (str[1] & 0x3f);
	for(i=0; i<sizeof(enumAktiveDetector); i++)
	{
	  if ( exp != enumAktiveDetector[i].index ) continue;
	  pText = enumAktiveDetector[i].detector;
	  break;
	}
	if ( i == sizeof(enumAktiveDetector) ) pText = "not found Detektor";
//	  *(short *)pReq->pInData = (short)(str[1] & 0x3f);
  }
  if ( strlen(pText) > pReq->inLength ) strncpy( pReq->pInData, pText, pReq->inLength);
  else strcpy(pReq->pInData, pText);
  return 1;
}

int getReceivedData(CdiRequestInfoBlk *pReq, char *str, unsigned int leng)
{
  if ( get6150ADdatachar(pReq, str, leng) ) return 1;

  if ( pReq->inType == cdiChar )				// string input	*/
  {
	//cdilog("rs232String: cdiChar check: length: %d; string: %s\n",leng, str);
	if ( pReq->conditionFlag == maskChangeWindow || pReq->conditionFlag == maskChangePercent )
	{
	  if ( pReq->requestCounter == 1 ) pReq->dataReturnCondition = DATA_PASS_FILTER; /* first call of Async or sync */
      else if ( pReq->patternFilter == NULL ) pReq->dataReturnCondition = DATA_PASS_FILTER; /* no filter => any data */
      else if ( strcmp(str,pReq->pInData) != 0 ) pReq->dataReturnCondition = DATA_PASS_FILTER; /* data changed compare to last one */
	}
	else /* only wait for specific pattern!! */
    {
	  /*if ( strcmp(str, pReq->patternFilter) == 0 ) */
	  if (pReq->patternFilter[0] == 0) pReq->dataReturnCondition = DATA_PASS_FILTER;
      if ( strstr(str, (char *)pReq->patternFilter) != NULL )
      {
		if ( pReq->conditionFlag == maskCompEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
	    /* else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK; */
	  }
      else if ( pReq->conditionFlag == maskCompNotEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
      /* else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK; */
	}
  }
  else
  {
	checkReturnCondition(pReq, pReq->pInData, str, leng, 1);
  }
  if ( pReq->dataReturnCondition == 0 ) return 0;
  /*
  if (NGdebug) dbglog("rs232 return condition OK: %s",str);
  */
	    //cdilog("rs232String: passt check read data: length: %d; string: %s, inLength: %d\n",leng, str, pReq->inLength);

  if ( leng > pReq->inLength ) strncpy( pReq->pInData, str, pReq->inLength);
  else strcpy(pReq->pInData, str);

  return 1;
}

void rs232String(int cdiLine, char *str, unsigned int leng)
{ 
  CdiLineInfoBlk *busHandle;
  /* CdiLineInfoBlk *pLine = busHandle[cdiLine]; */
  CdiRequestInfoBlk *pReq;
  /*
  if (NGdebug) 
    dbglog("call Back: cdi line: %d, port number: %d, string: %s, length: %d",
            cdiLine, busHandle[cdiLine]->lineNumber, str, leng);
  */
  if ((busHandle=cdiGetBusHandle(cdiLine)) == NULL)
  {
    cdilog("rs232String: Wrong cdi_line: %d for rs232String() call\n",cdiLine);
    return;
  }

  for ( busHandle->readCounter=0; busHandle->readCounter < busHandle->numberActiveRequest; busHandle->readCounter++ )
  {
    if ( busHandle->pReqList[ busHandle->readCounter ].active != recvRequest &&
    busHandle->pReqList[ busHandle->readCounter ].active != recvRequest2nd ) continue;
    pReq = &busHandle->pReqList[ busHandle->readCounter ];

	    //cdilog("rs232String: check read data: length: %d; string: %s\n",leng, str);

	if ( getReceivedData(pReq, str, leng) )
	{
	  cdiWaitSemaphore(&pReq->exec_SemId,0);
	  cdiExecRequest(pReq, readTaskIndex);
	  cdiPostSemaphore(&pReq->exec_SemId);
	}
  }
}

/****************************************/
int commTimeoutFcn(CdiRequestInfoBlk *pReq, int set, int taskIndex)
{
  SER_STRUCT *pSer = &serInfo[ ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1 ];
  if ( set == 0 ) return 0;
  if ( strcmp(pReq->pDev->lngName, ad6150Name) != 0 ) return 0;
  if ( pSer->tmoutStemp == 0 || abs(getTimeMilliseconds() - pSer->tmoutStemp) > resetInterval )
  {
	pSer->tmoutStemp = getTimeMilliseconds();
	pSer->tmoutCounter = 1;
  }
  else pSer->tmoutCounter++;

  return 0;
  if ( abs(getTimeMilliseconds() - pSer->tmoutStemp) > tmoutInterval )
  {
	pSer->tmoutStemp = 0;
	cdilog("exit after timeouts!\n");
	exit(0);
  }
  return 0;
}

/****************************************/
/*	send String							*/
/****************************************/
unsigned int sendSerialPort(SER_STRUCT *pSer, char *psOutput)
{
#ifndef CDI_COMM_LINUX
  if ( pSer->serFd == NULL )
#else
  if ( pSer->serFd < 0 )
#endif
  {
    cdilog(" port is not open yet! port: com%d\n", pSer->channel+1);
    return -file_not_open;
  }

#ifndef CDI_COMM_LINUX
  pSer->wrtRet = ResetEvent(pSer->overlap[1].hEvent);
  pSer->wrtRet = WriteFile(pSer->serFd, // handle to file
    (LPCVOID)psOutput,          // data buffer
    (DWORD)strlen(psOutput),    // number of bytes to write
    (LPDWORD)&pSer->iOut,             // number of bytes written
    (LPOVERLAPPED)&pSer->overlap[1]          // overlapped buffer
  );
  if ( pSer->wrtRet == 0 && GetLastError() == ERROR_IO_PENDING )
  {
   	pSer->wrtRet = WaitForSingleObject (pSer->overlap[1].hEvent, INFINITE);
    if ( pSer->wrtRet != WAIT_OBJECT_0 )
    {
      cdilog("WriteFile error! rs232 port: com%d; GetLastError(): 0x%x\n",pSer->channel+1, pSer->wrtRet);
      return -rs232_error;
    }
    pSer->wrtRet = GetOverlappedResult(
  			pSer->serFd,				// Handle to device
	  		&pSer->overlap[1],	// pointer to overlapped structure  
		  	&pSer->iOut,// pointer to actual bytes count  
  			FALSE );// wait flag!
    if ( pSer->wrtRet == 0 )
    {
      cdilog("WriteFile error! rs232 port: com%d; GetLastError(): 0x%x\n",pSer->channel+1, GetLastError());
      return -rs232_error;
    }
  }
  else
  {
    pSer->wrtRet = GetLastError();
    cdilog("WriteFile error! rs232 port: com%d; GetLastError(): 0x%x\n",pSer->channel+1, pSer->wrtRet);
    return -rs232_error;
  }
#else
  pSer->iOut = write(pSer->serFd, psOutput, strlen(psOutput));
  if (pSer->iOut < 0)
  {
    cdilog("write serial port(com%d) error %d %s\n", pSer->channel+1, errno, strerror(errno));
    return -rs232_error;
  }
  else
  {
//    printf("wrote %d chars: %s on serial port: com%d\n", pSer->iOut, psOutput, pSer->channel+1);
  }
#endif

  if (pSer->iOut < strlen(psOutput))
  {
    cdilog("write data to serial port error: written: %d; requested: %d", pSer->iOut, strlen(psOutput));
  }
  else 
  {
    if (busDebugLevel) dbglog("write to com%d ==> wrote %d chars: %s", pSer->channel+1, pSer->iOut, psOutput);
  }
  return pSer->iOut;
}

/****************************************************************************/
/* send request! input active flag = transRequest or transRequest2nd */
/* after, set active flag to finishRequest, recvRequest, or recvRequest2nd */
/****************************************************************************/
CDI_COMM_EXPORT void execRS232(CdiRequestInfoBlk *pReq)
{
  SER_STRUCT *pSer = &serInfo[ ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1 ];
  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }
  if ( pReq->active == transRequest ) /* start first transfer */
  {
    if ( pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag ) /* write first */
    {
      if ( sendSerialPort(pSer, pReq->pOutData) != pReq->outLength )
      {
        pReq->errorCode = rs232_error; /* write failed */
        if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
        pReq->active = finishRequest;
      } /* finish or 2ndReceive */
      else pReq->active = pReq->accessFlag == cdiWriteFlag ? finishRequest : recvRequest2nd;
    }
    else
	{
	  if ( abs(getTimeMilliseconds() - pSer->timeStemp) < pSer->interval )
	  {
		if ( getReceivedData(pReq, pSer->rcvData, pSer->recLength) )
		{
	      pReq->active = finishRequest;
		  return;
		}
	  }
	  pReq->active = recvRequest; /* jump to receive */
	}
    return;
  }
  /* second transfer */
  if ( pReq->accessFlag == cdiReadWriteFlag ) /* 2nd write */
  {
    if ( sendSerialPort(pSer, pReq->pOutData) != pReq->outLength )
    {
      pReq->errorCode = rs232_error;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    }
    pReq->active = finishRequest;
  }

  return;
}

/********************************************/
/*	receive string from port				*/
/********************************************/
//int receiveSerialPortString(int channelNumber, char *pString)
#ifndef CDI_COMM_LINUX
DWORD WINAPI GetSerialStringThread(SER_STRUCT *pSer)
#else
void *GetSerialStringThread(void *param)
#endif
{
  do
  {
#ifndef CDI_COMM_LINUX
	pSer->ret = ResetEvent(pSer->overlap[0].hEvent);
    pSer->ret = ReadFile(pSer->serFd,   // handle to file
                    (LPVOID)pSer->buf,  // data buffer
                    STRING_LENGTH,             // number of bytes to read
                    (LPDWORD)&pSer->length,    // number of bytes read
                    &pSer->overlap[0]           // overlapped buffer
                  );
	if ( pSer->ret == 0 && GetLastError() == ERROR_IO_PENDING )
    {
	  pSer->ret = WaitForSingleObject(pSer->overlap[0].hEvent, INFINITE);
      if ( pSer->ret != WAIT_OBJECT_0 )
      {
        cdilog("ReadFile(WaitForSingleObject) failed with error 0x%x; on port com%d; read length: %d\n", GetLastError(), pSer->channel+1, pSer->length);
        continue;
      }
      pSer->ret = GetOverlappedResult(
  			pSer->serFd,				// Handle to device
	  		&pSer->overlap[0],	// pointer to overlapped structure  
		  	&pSer->length,      // pointer to actual bytes count  
  			FALSE );            // wait flag!
      if ( pSer->ret == 0 )
      {
        cdilog("ReadFile(GetOverlappedResult) failed with error 0x%x; on port com%d; read length: %d\n", GetLastError(), pSer->channel+1, pSer->length);
        continue;
      }
    }
    else
    {
      cdilog("ReadFile Call failed with error 0x%x; on port com%d; read length: %d\n", GetLastError(), pSer->channel+1, pSer->length);
      continue;
    }

    if ( pSer->length <= 0 )
    {
      cdilog("ReadFile read wrong data, length; on port com%d; read length: %d\n", pSer->channel+1, pSer->length);
      continue;
    }
#else
	SER_STRUCT *pSer = (SER_STRUCT *)param;
  	pSer->ret = read(pSer->serFd, pSer->buf, STRING_LENGTH);
	if ( pSer->canonicalMode )
  	{
	  pSer->buf[pSer->ret] = 0;
	  cdilog(" comport: com%d; canonicalMode input string: %s\n", pSer->channel+1, pSer->buf);
	  pSer->length = pSer->ret;
	}
  	else
	{
	  pSer->length = pSer->ret;
	  for ( ; pSer->ret > 0 && pSer->length < STRING_LENGTH; )
	  {
		FD_ZERO( &(pSer->fdset) );
		FD_SET( pSer->serFd, &(pSer->fdset) );    

    	pSer->tout.tv_sec = (pSer->usOneChar * 20) / 1000000;		// wait time for 20 char transmission
		pSer->tout.tv_usec = (pSer->usOneChar * 20) % 1000000;   /* convert to uSeconds */

    	pSer->ret = select( pSer->serFd + 1, &(pSer->fdset), NULL, NULL, &(pSer->tout) );
		if ( pSer->ret > 0 )	// still data there, no time out
		{
		  pSer->ret = read(pSer->serFd, &(pSer->buf[pSer->length]), STRING_LENGTH - pSer->length);
		  if ( pSer->ret > 0 ) pSer->length += pSer->ret;
    	}
		if ( pSer->length >= STRING_LENGTH )
        {
          cdilog("String overflow!!!!!!!!! length: %d(received), %d(maximum); on serial port: com%d;\n",pSer->length, STRING_LENGTH, pSer->channel+1);
          pSer->length = STRING_LENGTH;
        }
        //	if( select( serFd + 1, &fdset, NULL, NULL, &tout ) > 0 )  return read( serFd, c, 1 ) == 1 ? 0 : -1;
      }
  	}
#endif
    pSer->buf[pSer->length] = 0;
    pSer->timeStemp = getTimeMilliseconds();
	pSer->tmoutStemp = 0;
	pSer->tmoutCounter = 0;
	memcpy(pSer->rcvData, pSer->buf, pSer->length + 1);
	pSer->recLength = pSer->length;
    if (busDebugLevel) dbglog("comport: com%d; input string: %s", pSer->channel+1, pSer->buf);
    //cdilog("comport: com%d; input length: %d; string: %s\n", pSer->channel+1, pSer->length, pSer->buf);
	if (pSer->receiveString) pSer->receiveString(pSer->cdiLine, pSer->buf, pSer->length);
  } while (1);
}

/****************************************/
/* open port 0-3(com1 to com4)			*/
/* parameters[], see header file		*/
/****************************************/
int openSerialPort(int channelNumber, int *parameters, void (*callBack)(int, char*, unsigned int), int cdiLine)
{
#ifndef CDI_COMM_LINUX
  char pcCommPort[16];
  HANDLE hCom;
  DCB dcb;
  COMMTIMEOUTS commtmouts;
  BOOL fSuccess;
  DWORD dwCdiThreadId;
#else
  char devName[16];
  struct termios tio;
#endif
  SER_STRUCT *pSer;
  int i;

  if ( channelNumber < 1 || channelNumber > N_SERIAL_LINES ) return -illegal_line;
  channelNumber--;

  // find correct baudrate
  for (i=0; i<numBaudRate; i++) if (parameters[paramBaudrate] == baudRate[i].inputBaud) break;
  if ( i == numBaudRate )
  {
    cdilog("input wrong! baudrate: %d for Com%d!\n", parameters[paramBaudrate], channelNumber+1);
	return -invalid_data;
  }

  pSer = &serInfo[channelNumber];
#ifndef CDI_COMM_LINUX
  if ( pSer->serFd != NULL )
#else
  if ( pSer->serFd >= 0 )
#endif
  {
    cdilog("serial already opened! line: %d! for Com%d\n", channelNumber, channelNumber+1);
    return 0;
  }

#ifndef CDI_COMM_LINUX
  // create handle
  sprintf( pcCommPort, "COM%d", channelNumber+1 );
  printf("serialPortName=%s\n", pcCommPort);
  cdilog("serialPortName=%s\n", pcCommPort);
  hCom = CreateFile( pcCommPort,
                     GENERIC_READ | GENERIC_WRITE,
                     0,    // comm devices must be opened w/exclusive-access
                     NULL, // no security attributes
                     OPEN_EXISTING, // comm devices must use OPEN_EXISTING
                     FILE_FLAG_OVERLAPPED,    // not overlapped I/O
                     NULL  // hTemplate must be NULL for comm devices
                     );

  if (hCom == INVALID_HANDLE_VALUE) {      // Handle the error.
    cdilog("CreateFile for %s failed with error 0x%x.\n", pcCommPort, GetLastError());
    return -file_not_open;
  }

  // set buffer size
  fSuccess = SetupComm(
    hCom,           // handle to communications device
    dwInQueue,      // size of input buffer
    dwOutQueue      // size of output buffer
  );
  if (!fSuccess) {      // Handle the error.
    cdilog("SetupComm for %s failed with error %d.\n", pcCommPort, GetLastError());
    CloseHandle(hCom);
    return -rs232_error;
  }

  // We will build on the current configuration, and skip setting the size
  // of the input and output buffers with SetupComm.
  fSuccess = GetCommState(hCom, &dcb);
  if (!fSuccess) {      // Handle the error.
    cdilog("GetCommState for %s failed with error %d.\n", pcCommPort, GetLastError());
    CloseHandle(hCom);
    return -rs232_error;
  }
  
  memcpy(&pSer->dcbCopy, &dcb, sizeof(DCB));

  // Fill in the DCB: baud=57,600 bps, 8 data bits, no parity, and 1 stop bit.
  dcb.BaudRate = parameters[paramBaudrate];;     // set the baud rate
  dcb.ByteSize = parameters[param8bits] == 0 ? 7 : 8;           // data size, xmit, and rcv
  dcb.fParity = parameters[paramParity] ? TRUE : FALSE;         // no parity bit
  dcb.StopBits = parameters[param2StopBits] ? TWOSTOPBITS : ONESTOPBIT;    // one stop bit
  if ( dcb.fParity ) dcb.Parity = parameters[paramOddParity] ? ODDPARITY : EVENPARITY;
  else dcb.Parity = NOPARITY;
  if ( parameters[paramFlowControl] )
  {
    dcb.fRtsControl = RTS_CONTROL_ENABLE;
    dcb.fOutxCtsFlow = TRUE;
  }

  fSuccess = SetCommState(hCom, &dcb);
  if (!fSuccess) {     // Handle the error.
    cdilog("SetCommState for %s failed with error %d.\n", pcCommPort, GetLastError());
    CloseHandle(hCom);
    return -rs232_error;
  }

  if ( (fSuccess = GetCommTimeouts(hCom, &commtmouts)) == 0 )
  {      // Handle the error.
    cdilog("GetCommTimeouts for %s failed with error %d.\n", pcCommPort, GetLastError());
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    CloseHandle(hCom);
    return -rs232_error;
  }
  memcpy(&pSer->commtmoutsCopy, &commtmouts, sizeof(COMMTIMEOUTS));
  commtmouts.ReadIntervalTimeout = 1000 * 10 * 20 / parameters[paramBaudrate];  // timeout in ms for 10 char transmission, 10 bit per char
  commtmouts.ReadTotalTimeoutConstant = 0;  // timeout not used!
  commtmouts.ReadTotalTimeoutMultiplier = 0;
//  commtmouts.WriteTotalTimeoutConstant = 1;
//  commtmouts.WriteTotalTimeoutMultiplier = 1;
  if ( (fSuccess = SetCommTimeouts(hCom, &commtmouts)) == 0 )
  {     // Handle the error.
    cdilog("SetCommTimeouts for %s failed with error %d.\n", pcCommPort, GetLastError());
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    CloseHandle(hCom);
    return -rs232_error;
  }

  // create event for read and write
  for (i=0; i<2; i++)
  {
    pSer->overlap[i].Offset = 0;
	pSer->overlap[i].OffsetHigh = 0;
    pSer->overlap[i].hEvent = CreateEvent(
						NULL,	// pointer to security attributes
						TRUE,	// flag for manual-reset event
						FALSE,	// flag for initial state(false: nonsignaled)
						NULL );	// pointer to event-object name
    if ( pSer->overlap[i].hEvent == NULL ) break;
  }
  if ( i < 2 )
  {
    if ( i == 1 ) CloseHandle( pSer->overlap[i-1].hEvent );
    cdilog("CreateEvent for %s failed with error %d.\n", pcCommPort, GetLastError());
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    fSuccess = SetCommTimeouts(hCom, &pSer->commtmoutsCopy);
    CloseHandle(hCom);
  	return -rs232_error;
  }

	// Waiting for messages and work them out on master module
  if ( (pSer->tIdMaster = CreateThread(NULL,0,GetSerialStringThread,(void *)pSer,0,&dwCdiThreadId)) == NULL )
  {
	cdilog ("CreateThread line: %s error",pcCommPort);
    fSuccess = SetCommState(hCom, &(pSer->dcbCopy));
    fSuccess = SetCommTimeouts(hCom, &pSer->commtmoutsCopy);
    for ( i=0; i<2; i++ ) CloseHandle( pSer->overlap[i].hEvent );
    CloseHandle(hCom);
	return -un_allocated;
  }
#else
  if (parameters[bluetoothDevice] == 0)
   sprintf( devName, "/dev/ttyS%d", channelNumber );
  else
   sprintf( devName, "/dev/rfcomm%d", channelNumber );
   printf("serialPortName=%s\n", devName);
   cdilog("serialPortName=%s\n", devName);

   // open as read, write and not controlling terminal
   if( ( pSer->serFd = open( devName, O_RDWR | O_NOCTTY ) ) < 0 )
   {
  	 printf("open %s port (com%d) error %d %s\n", devName, channelNumber+1, errno, strerror(errno));
  	 cdilog("open %s port (com%d) error %d %s\n", devName, channelNumber+1, errno, strerror(errno));
	 return -file_not_open;
   }

   tcgetattr( pSer->serFd, &(pSer->oldtio) );
   memset(  &tio, '\0', sizeof(tio) );

   // set control modes

   /* tio.c_cflag = B38400 | CRTSCTS | CS8 | CLOCAL | CREAD; */
   // bartrate, RTSc control, 7 data bits, 2 stop bits, even parity, no modem, enable receiver 
   //   tio.c_cflag = B9600 | CRTSCTS | CS7 | CSTOPB | PARENB | CLOCAL | CREAD; 
   
   // baudRate, no modem control, enable read
   tio.c_cflag = baudRate[i].defBaud | CLOCAL | CREAD | (parameters[param8bits] == 0 ? CS7 : CS8);
   if ( parameters[param2StopBits] ) tio.c_cflag |= CSTOPB;
   if ( parameters[paramParity] ) tio.c_cflag |= PARENB;
   if ( parameters[paramOddParity] ) tio.c_cflag |= (PARENB | PARODD);
   if ( parameters[paramFlowControl]) tio.c_cflag |= CRTSCTS;

   // input modes
   tio.c_iflag = IGNPAR;	/* ignore bytes with framing and parity error */
   if ( parameters[paramCanonicalInput] ) tio.c_iflag |= ICRNL;

   // output modes
   tio.c_oflag = 0;			// raw output

   // local modes
   tio.c_lflag = parameters[paramCanonicalInput] ? ICANON : 0;/* non-canonical input mode */

   // control chars
   tio.c_cc[VTIME]  = 0;  /* do not use inter-character timer */
   tio.c_cc[VMIN]   = 1;  /* blocking read until 1 character received */
  
   tcflush( pSer->serFd, TCIOFLUSH );
   tcsetattr( pSer->serFd, TCSANOW, &tio );

	// Waiting for messages and work them out on master module
   if ( pthread_create( &(pSer->tIdMaster), NULL, GetSerialStringThread, (void *)pSer ) )
   {
	  cdilog ("pthread_create error for serial port: com%d",channelNumber+1);
	  tcflush( pSer->serFd, TCIOFLUSH );
	  tcsetattr( pSer->serFd, TCSANOW, &(pSer->oldtio) );
	  close( pSer->serFd );
	  pSer->serFd = -1;
	  return -un_allocated;
   }
#endif
  pSer->baudRate = parameters[paramBaudrate];
  pSer->usOneChar = 1000000 / pSer->baudRate * 10;	// 10 bit per char
  pSer->channel = channelNumber;
  pSer->cdiLine = cdiLine;
  pSer->receiveString = callBack;
  pSer->canonicalMode = parameters[paramCanonicalInput];
  pSer->timeStemp = 0;
  pSer->interval = parameters[dataInterval] <= 0 ? 500 : parameters[dataInterval];

#ifndef CDI_COMM_LINUX
  pSer->serFd = hCom;
  fSuccess = PurgeComm(pSer->serFd, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
  if ( fSuccess == 0 ) cdilog("PurgeComm error! GetLastError(): 0x%x\n",GetLastError());
  cdilog("Serial port %s successfully reconfigured. handle: %d\n", pcCommPort, pSer->serFd);
#else
  cdilog("Serial port %s successfully reconfigured. handle: %d\n", devName, pSer->serFd);
#endif

  return 0;
}

/****************************************/
#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}
CDI_COMM_EXPORT int initRS232ForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  int parameter[NUMBER_EXEC_PARAM];
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
  return openSerialPort(busLine, parameter, rs232String, cdiLine);
}

/****************************************/
/*	Close								*/
/****************************************/
CDI_COMM_EXPORT int closeSerialPort(int channelNumber)
{ SER_STRUCT *pSer;
#ifndef CDI_COMM_LINUX
  int i;
  BOOL fSuccess;
#endif
  if ( channelNumber < 1 || channelNumber > N_SERIAL_LINES ) return -illegal_line;

  cdilog("close serialPort! channelNumber: Com-%d\n", channelNumber);

  pSer = &serInfo[--channelNumber];
#ifndef CDI_COMM_LINUX
  if ( pSer->serFd == NULL ) return -file_not_open;
  CancelIo(pSer->serFd);
  CloseHandle( pSer->tIdMaster );
  fSuccess = SetCommState(pSer->serFd, &pSer->dcbCopy);
  fSuccess = GetCommTimeouts(pSer->serFd, &pSer->commtmoutsCopy);
  for(i=0;i<2;i++) CloseHandle( pSer->overlap[i].hEvent );
  CloseHandle(pSer->serFd);
  pSer->serFd = NULL;
#else
  pthread_cancel( pSer->tIdMaster );
  tcflush( pSer->serFd, TCIOFLUSH );
  tcsetattr( pSer->serFd, TCSANOW, &(pSer->oldtio) );
  close( pSer->serFd );
  pSer->serFd = -1;
#endif
  return 0;
}
