#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "toolkit.h"
#include "tine.h"
#include "cdi.h"
#include "tinecdi.h"

#ifdef CDICLBRFCN_EXPORTS
# define CDICLBRFCN_EXPORT __declspec(dllexport)
#else
# define CDICLBRFCN_EXPORT 
#endif

// 3 intervals
double inv[3] = {0.4, 9.4, 7.5};
double prf[5] = {1.8, 0.47, 1.8, 0.65, -1.708};
double zpr[5] = {-3.6, -2.7, -15, -9.6, 0.15};

CDICLBRFCN_EXPORT double vdDruck(double v)
{ int i;
  double logp, spannung = v / 409.5;    // 0-10 volt for 12 Bits(0-4095)
  if ( spannung > 10 ) return 0;
  for (i=0; i<2; i++) if (spannung < inv[i]) break;
  logp = prf[i]* spannung + zpr[i];
  return pow(10,logp);
}

CDICLBRFCN_EXPORT double pdDruck(double v)
{ int i;
  double logp, spannung = v / 409.5;    // 0-10 volt for 12 Bits(0-4095)
  if ( spannung > 10 ) return 0;
  i = spannung > inv[2] ? 4 : 3;
  logp = prf[i] * spannung + zpr[i];
  if ( i == 4 ) logp += 0.14 * spannung * spannung; // include non linear part
  return pow(10,logp);
}

CDICLBRFCN_EXPORT double bit12sgn(double v)
{
  short w = (short)v;
  short sgn = sgn = (w & 0x0800) == 0 ? 1 : -1;
  w &= 0x07ff;
  w *= sgn;
  return (double)w;
}

CDICLBRFCN_EXPORT double nib2exp(double v)
{
  short w = (short)v;
  return (double)((w&0xfff) * pow(10,-((short)(w>>12))));
}

CDICLBRFCN_EXPORT double bit12Recv(double value)
{
	short current = (short) value;

	// NOTE: in the preceding version this value was set to 0x7ff which means that 
	// the 12th bit won't be considered! These seems not to work with this function, 
	// because it will only recognize negativ values but not positiv ones!
	current &= 0x0fff; // only recognize the lower 12 bits
	current -= 2047; // -2047 is equal to -0x07ff

	return (double) current;
}

CDICLBRFCN_EXPORT double bit12Send(double value)
{
	short current = (short) value;
	current += 2047;
	return (double) current;
}

void cdiRegisterFunctions(void)
{
  cdi_RegisterCalibrationFunction("vdDruck",vdDruck);
  cdi_RegisterCalibrationFunction("pdDruck",pdDruck);
  cdi_RegisterCalibrationFunction("bit12sgn",bit12sgn);
  cdi_RegisterCalibrationFunction("nib2exp",nib2exp);
  cdi_RegisterCalibrationFunction("bit12Recv",bit12Recv);
  cdi_RegisterCalibrationFunction("bit12Send",bit12Send);
}

BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      cdiRegisterFunctions();
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      break;
  }
  return (TRUE);
}
